"""
This module contains the implementation of the
FR Test Statistic based estimate and the
FNN based estimate for the Bayes Error Rate
"""
from abc import abstractmethod
from typing import Optional, Tuple, Union

import numpy as np
import torch
from scipy.sparse import coo_matrix, csr_matrix
from scipy.sparse.csgraph import minimum_spanning_tree
from scipy.spatial.distance import pdist, squareform

from daml._internal.datasets.datasets import DamlDataset
from daml._internal.metrics.aria.base import _AriaMetric
from daml._internal.metrics.outputs import BEROutput

from .utils import permute_to_numpy, permute_to_torch


class _MultiClassBer(_AriaMetric):
    def __init__(
        self, encode: bool = False, device: Union[str, torch.device] = "cpu"
    ) -> None:
        """Constructor method"""

        super().__init__(encode, device)

    @abstractmethod
    def _multiclass_ber(
        self,
        X: np.ndarray,
        y: np.ndarray,
    ) -> Tuple[float, float]:
        """Abstract method for the implementation of multiclass BER calculation"""

    def _get_classes_counts(self, labels: np.ndarray) -> Tuple[int, np.intp]:
        classes, counts = np.unique(labels, return_counts=True)
        M = len(classes)
        if M < 2:
            raise ValueError("Label vector contains less than 2 classes!")
        N = np.sum(counts)
        return M, N

    def evaluate(
        self, dataset: DamlDataset, encode: Optional[bool] = None
    ) -> BEROutput:
        """
        Return the Bayes Error Rate estimate

        Parameters
        ----------
        dataset : DamlDataset
            Dataset containing (n_samples x n_features) array of (padded) instance
            embeddings and n_samples vector of class labels with M unique classes.
        Returns
        -------
        BEROutput
            The estimated upper and lower bounds of the Bayes Error Rate

        Raises
        ------
        ValueError
            If unique classes M < 2
        """
        X: np.ndarray = dataset.images
        y: np.ndarray = dataset.labels

        # Parameter encode can override class encode
        do_encode = self.encode if encode is None else encode
        # Pass X through an autoencoder before evaluating BER
        if do_encode:
            if not self._is_trained or self.model is None:
                raise TypeError(
                    "Tried to encode data without fitting a model.\
                    Try calling Metric.fit_dataset(dataset) first."
                )
            else:
                images = X if isinstance(X, torch.Tensor) else permute_to_torch(X)
                embeddings = self.model.encode(images).detach().cpu().numpy()
        else:
            embeddings = X if isinstance(X, np.ndarray) else permute_to_numpy(X)

        assert isinstance(embeddings, np.ndarray)
        ber, ber_lower = self._multiclass_ber(embeddings, y)
        return BEROutput(ber=ber, ber_lower=ber_lower)


class MultiClassBerMST(_MultiClassBer):
    """
    Implements the FR Test Statistic based estimator for the Bayes Error Rate

    Note
    ----
    `Learning to Bound the Multi-class Bayes Error (Th. 3 and Th. 4) <https://arxiv.org/abs/1811.06419>`_
    """  # noqa F401

    def _multiclass_ber(
        self,
        X: np.ndarray,
        y: np.ndarray,
    ) -> Tuple[float, float]:
        """
        Calculates the Bayes Error Rate estimate

        Parameters
        ----------
        X : np.ndarray
            (n_samples x n_features) array of covariates (or image embeddings)
        y : np.ndarray
            n_samples vector of class labels with M unique classes. 2 <= M <= 10

        Returns
        -------
        float
            Estimate of the Bayes Error Rate

        Raises
        ------
        ValueError
            If unique classes M < 2 or M > 10
        """
        M, N = self._get_classes_counts(y)

        # All features belong on second dimension
        X = X.reshape((X.shape[0], -1))
        # We add a small constant to the distance matrix to ensure scipy interprets
        # the input graph as fully-connected.
        dense_eudist = squareform(pdist(X)) + 1e-4
        eudist_csr = csr_matrix(dense_eudist)
        tree = coo_matrix(minimum_spanning_tree(eudist_csr))
        sum = np.sum([y[tree.row[i]] != y[tree.col[i]] for i in range(N - 1)])
        deltas = sum / (2 * N)
        upper = 2 * deltas
        lower = ((M - 1) / (M)) * (1 - (1 - 2 * ((M) / (M - 1)) * deltas) ** 0.5)
        return upper, lower


class MultiClassBerFNN(_MultiClassBer):
    """
    Implements the KNN Test Statistic based estimator for the Bayes Error Rate

    Parameters
    ----------
    X : np.ndarray
        (n_samples x n_features) array of covariates (or image embeddings)
    y : np.ndarray
        n_samples vector of class labels with M unique classes. 2 <= M <= 10

    Returns
    -------
    float
        Estimate of the Bayes Error Rate

    Raises
    ------
    ValueError
        If unique classes M < 2 or M > 10

    See Also
    --------
    `Learning to Bound the Multi-class Bayes Error (Th. 3 and Th. 4) <https://arxiv.org/abs/1811.06419>`_
    """  # noqa F401

    def _multiclass_ber(
        self,
        X: np.ndarray,
        y: np.ndarray,
    ) -> Tuple[float, float]:
        M, N = self._get_classes_counts(y)

        # All features belong on second dimension
        X = X.reshape((X.shape[0], -1))
        nn_indices = self._compute_neighbors(X, X)
        deltas = float(np.count_nonzero(y[nn_indices] - y) / (2 * N))
        upper = 2 * deltas
        lower = ((M - 1) / (M)) * (1 - (1 - 2 * ((M) / (M - 1)) * deltas) ** 0.5)
        return upper, lower



"""
This module contains the implementation of the
FR Test Statistic based estimate for the upperbound
average precision (UAP_MST) and empirical mean
precision (UAP_EMP)
"""
from typing import Tuple

import numpy as np
import torch
from scipy.sparse import csr_matrix
from scipy.sparse.csgraph import minimum_spanning_tree
from scipy.spatial.distance import pdist, squareform
from sklearn.metrics import average_precision_score

from daml._internal.datasets.datasets import DamlDataset
from daml._internal.metrics.aria.base import _AriaMetric
from daml._internal.metrics.outputs import UAPOutput

from .utils import permute_to_numpy, permute_to_torch


class UAP_MST(_AriaMetric):
    def __init__(self, encode: bool = False) -> None:
        """Constructor method"""

        super().__init__(encode, device="cpu")

    def _get_classes_counts(self, labels: np.ndarray) -> Tuple[int, np.intp]:
        classes, counts = np.unique(labels, return_counts=True)
        M = len(classes)
        N = np.sum(counts)
        return M, N

    def evaluate(self, dataset: DamlDataset) -> UAPOutput:
        """
        Return the Upperbound Average Precision estimate

        Parameters
        ----------
        dataset : DamlDataset
            Dataset containing (n_samples x n_features) array of (padded) instance
            embeddings and n_samples vector of class labels with M unique classes.

        Returns
        -------
        UAPOutput
            The estimated UAP
        """
        X: np.ndarray = dataset.images
        y: np.ndarray = dataset.labels

        # If self.encode == True, pass X through an autoencoder before evaluating BER
        if self.encode:
            if not self._is_trained or self.model is None:
                raise TypeError(
                    "Tried to encode data without fitting a model.\
                    Try calling Metric.fit_dataset(dataset) first."
                )
            else:
                images = X if isinstance(X, torch.Tensor) else permute_to_torch(X)
                embeddings = self.model.encode(images).numpy()
        else:
            embeddings = X if isinstance(X, np.ndarray) else permute_to_numpy(X)

        assert isinstance(embeddings, np.ndarray)
        uap = self._uap(embeddings, y)
        return UAPOutput(uap=uap)

    def _uap(
        self,
        X: np.ndarray,
        y: np.ndarray,
    ) -> float:
        """
        Calculates the UAP estimate

        Parameters
        ----------
        X : np.ndarray
            (n_samples x n_features) array of (padded) instance embeddings
        y : np.ndarray
            n_samples vector of class labels with M unique classes. 2 <= M <= 10

        Returns
        -------
        float
            Estimate of the UAP

        Raises
        ------
        ValueError
            If unique classes M < 2 or M > 10
        """
        M, N = self._get_classes_counts(y)

        # All features belong on second dimension
        X = X.reshape((X.shape[0], -1))
        # We add a small constant to the distance matrix to ensure scipy interprets
        # the input graph as fully-connected.
        dense_eudist = squareform(pdist(X)) + 1e-4
        eudist_csr = csr_matrix(dense_eudist)
        tree = minimum_spanning_tree(eudist_csr).todense()
        conf_mat = np.zeros((M, M))
        for i in range(len(tree)):
            edges = np.where(tree[i, :] != 0)[1]
            for j in range(len(edges)):
                conf_mat[int(y[i]), int(y[edges[j]])] = (
                    conf_mat[int(y[i]), int(y[edges[j]])] + 1
                )
        probs = np.zeros((M, M))
        for i in range(M):
            probs[i, :] = conf_mat[i, :] / np.sum(conf_mat[i, :])
        lab_inds = np.zeros((N, M))
        prob_inds = np.zeros((N, M))
        for i in range(N):
            lab_inds[i, int(y[i])] = 1
            prob_inds[i, :] = probs[int(y[i]), :]
        uap = average_precision_score(lab_inds, prob_inds)
        return float(uap)


class UAP_EMP(_AriaMetric):
    def __init__(self) -> None:
        """Constructor method"""

        super().__init__(encode=False, device="cpu")

    def evaluate(self, dataset: DamlDataset, scores: np.ndarray) -> UAPOutput:
        """
        Return the Upper Average Precision estimate

        Parameters
        ----------
        dataset : DamlDataset
            Dataset containing (n_samples x n_features) array of (padded) instance
            embeddings and n_samples vector of class labels with M unique classes.

        scores : np.ndarray
            A 2D array of class probabilities per image

        Returns
        -------
        UAPOutput
            The estimated UAP
        """
        y: np.ndarray = dataset.labels

        uap = float(average_precision_score(y, scores, average="weighted"))
        return UAPOutput(uap=uap)



from abc import ABC
from typing import Literal, Optional, Union

import numpy as np
import torch
import torch.nn as nn
from sklearn.neighbors import NearestNeighbors

from daml._internal.datasets.datasets import DamlDataset
from daml._internal.models.pytorch.autoencoder import AERunner, AETrainer

from .utils import permute_to_torch


class _AriaMetric(ABC):
    """Abstract base class for ARiA metrics"""

    def __init__(self, encode: bool, device: Union[str, torch.device]):
        """Constructor method"""

        self.encode = encode
        self.model: Optional[AERunner] = None
        self._is_trained: bool = False
        self._device = device

    def fit_dataset(
        self,
        dataset: DamlDataset,
        epochs: int = 3,
        model: Optional[nn.Module] = None,
    ) -> None:
        """
        Trains a model on a dataset to be used during calculation of metrics.

        Parameters
        ----------
        dataset : DamlDataset
            An array of images for the model to train on
        epochs : int, default 3
            Number of epochs to train the detector for.

        """
        if model is None:
            images: torch.Tensor = permute_to_torch(dataset.images)
            self.model = AETrainer(model, images.shape[1], device=self._device)
            self.model.train(images, epochs)
        elif isinstance(model, nn.Module):
            self.model = AERunner(model, device=self._device)
        else:
            raise TypeError(f"Given model is of type {type(model)}, expected nn.Module")
        self._is_trained = True
        self.encode = True

    def _compute_neighbors(
        self,
        A: np.ndarray,
        B: np.ndarray,
        k: int = 1,
        algorithm: Literal["auto", "ball_tree", "kd_tree"] = "auto",
    ) -> np.ndarray:
        """
        For each sample in A, compute the nearest neighbor in B

        Parameters
        ----------
        A, B : np.ndarray
            The n_samples and n_features respectively
        k : int
            The number of neighbors to find
        algorithm : Literal
            Tree method for nearest neighbor (auto, ball_tree or kd_tree)

        Note
        ----
            Do not use kd_tree if n_features > 20

        Returns
        -------
        List:
            Closest points to each point in A and B

        See Also
        --------
        :func:`sklearn.neighbors.NearestNeighbors`
        """

        nbrs = NearestNeighbors(n_neighbors=k + 1, algorithm=algorithm).fit(B)
        nns = nbrs.kneighbors(A)[1]
        nns = nns[:, 1]

        return nns



import numpy as np
import torch


def pytorch_to_numpy(tensor: torch.Tensor) -> np.ndarray:
    """
    Converts a PyTorch tensor to a NumPy array
    """
    if isinstance(tensor, np.ndarray):  # Already array, return
        return tensor
    if not isinstance(tensor, torch.Tensor):
        raise TypeError("Tensor is not of type torch.Tensor")

    x: np.ndarray = tensor.detach().cpu().numpy()
    return x


def numpy_to_pytorch(array: np.ndarray) -> torch.Tensor:
    """
    Converts a NumPy array to a PyTorch tensor
    """
    if isinstance(array, torch.Tensor):  # Already tensor, return
        return array
    if not isinstance(array, np.ndarray):
        raise TypeError("Array is not of type numpy.ndarray")
    x: torch.Tensor = torch.from_numpy(array.astype(np.float32))
    return x


def permute_to_torch(array: np.ndarray) -> torch.Tensor:
    """
    Converts and permutes a NumPy image array into a PyTorch image tensor.

    Parameters
    ----------
    array: np.ndarray
        Array containing image data in the format NHWC

    Returns
    -------
    torch.Tensor
        Tensor containing image data in the format NCHW
    """
    x = numpy_to_pytorch(array)
    x = x.permute(0, 3, 1, 2)  # NHWC -> NCHW
    return x


def permute_to_numpy(tensor: torch.Tensor) -> np.ndarray:
    """
    Converts and permutes a PyTorch image tensor into a NumPy image array.

    Does not permute if given np.ndarray

    Parameters
    ----------
    tensor: torch.Tensor
        Tensor containing image data in the format NCHW

    Returns
    -------
    np.ndarray
        Array containing image data in the format NHWC
    """
    x = tensor.permute(0, 2, 3, 1)
    x = pytorch_to_numpy(x)  # NCHW -> NHWC
    return x



"""
This module contains the implementation of Dp Divergence
using the First Nearest Neighbor and Minimum Spanning Tree algorithms
"""
from abc import abstractmethod
from typing import Optional, Union

import numpy as np
import torch
from scipy.sparse import csr_matrix
from scipy.sparse.csgraph import minimum_spanning_tree
from scipy.spatial.distance import pdist, squareform

from daml._internal.datasets.datasets import DamlDataset
from daml._internal.metrics.aria.base import _AriaMetric
from daml._internal.metrics.outputs import DivergenceOutput

from .utils import permute_to_numpy, permute_to_torch


class _DpDivergence(_AriaMetric):
    """
    For more information about this divergence, its formal definition,
    and its associated estimators
    see https://arxiv.org/abs/1412.6534.
    """

    def __init__(
        self, encode: bool = False, device: Union[str, torch.device] = "cpu"
    ) -> None:
        """Constructor method"""
        super().__init__(encode, device)

    @abstractmethod
    def calculate_errors(self, data: np.ndarray, labels: np.ndarray) -> int:
        """Abstract method for the implementation of divergence calculation"""

    def create_encoding(self, imgs_1: torch.Tensor, imgs_2: torch.Tensor) -> np.ndarray:
        """Reduces image dimensions using a trained autoencoder and stacks results"""
        if not self._is_trained:
            raise ValueError(
                "Tried to encode data without fitting a model.\
                    Try calling Metric.fit_dataset(dataset) first."
            )
        if self.model is None:
            raise TypeError(
                "Model is None. Try calling Metric.fit_dataset(dataset) first."
            )

        # Pass inputs through model
        tensor_a = self.model.encode(imgs_1).flatten(start_dim=1)
        tensor_b = self.model.encode(imgs_2).flatten(start_dim=1)

        # Combine data into one dataset and return
        return torch.vstack((tensor_a, tensor_b)).detach().cpu().numpy()

    def evaluate(
        self,
        dataset_a: DamlDataset,
        dataset_b: DamlDataset,
        encode: Optional[bool] = None,
    ) -> DivergenceOutput:
        """
        Calculates the divergence and any errors between two datasets

        Parameters
        ----------
        dataset_a, dataset_b : DamlDataset
            Datasets to calculate the divergence between

        Returns
        -------
        DivergenceOutput
            Dataclass containing the dp divergence and errors during calculation

        Note
        ----
        A and B must be 2 dimensions, and equivalent in size on the second dimension
        """

        imgs_a: np.ndarray = dataset_a.images
        imgs_b: np.ndarray = dataset_b.images

        do_encode = self.encode if encode is None else encode
        # Pass dataset_a and dataset_b through an autoencoder
        # before evaluating dp divergence
        if do_encode:
            i1 = (
                imgs_a if isinstance(imgs_a, torch.Tensor) else permute_to_torch(imgs_a)
            )
            i2 = (
                imgs_b if isinstance(imgs_b, torch.Tensor) else permute_to_torch(imgs_b)
            )
            data = self.create_encoding(i1, i2)
        else:
            # Input could be a torch.Tensor (future update)
            imgs_a = (
                imgs_a if isinstance(imgs_a, np.ndarray) else permute_to_numpy(imgs_a)
            )
            imgs_b = (
                imgs_b if isinstance(imgs_b, np.ndarray) else permute_to_numpy(imgs_b)
            )

            # Combine data into one dataset
            data = np.vstack((imgs_a, imgs_b))

        N = imgs_a.shape[0]
        M = imgs_b.shape[0]
        labels = np.vstack([np.zeros([N, 1]), np.ones([M, 1])])
        errors = self.calculate_errors(data, labels)
        dp = 1 - ((M + N) / (2 * M * N)) * errors
        return DivergenceOutput(dpdivergence=dp, error=errors)


class DpDivergenceMST(_DpDivergence):
    """
    A minimum spanning tree implementation of dp divergence

    Warning
    -------
        MST is very slow in this implementation, this is unlike matlab where
        they have comparable speeds
        Overall, MST takes ~25x LONGER!!
        Source of slowdown:
        conversion to and from CSR format adds ~10% of the time diff between
        1nn and scipy mst function the remaining 90%
    """

    # TODO
    # - validate the input algorithm
    # - improve speed for MST, requires a fast mst implementation
    # mst is at least 10x slower than knn approach

    def calculate_errors(self, data: np.ndarray, labels: np.ndarray) -> int:
        """
        Returns the divergence between two datasets using a minimum spanning tree

        Parameters
        ----------
        data : np.ndarray
            Array containing images from two datasets
        labels : np.ndarray
            Array showing which dataset each image belonged to

        Returns
        -------
        int
            Number of edges connecting the two datasets

        Note
        ----
        We add a small constant to the distance matrix to ensure scipy interprets
        the input graph as fully-connected.
        """
        dense_eudist = squareform(pdist(data)) + 1e-4
        eudist_csr = csr_matrix(dense_eudist)
        mst = minimum_spanning_tree(eudist_csr)
        mst = mst.toarray()
        edgelist = np.transpose(np.nonzero(mst))
        errors = np.sum(labels[edgelist[:, 0]] != labels[edgelist[:, 1]])
        return errors


class DpDivergenceFNN(_DpDivergence):
    """
    A first nearest neighbor implementation of dp divergence
    """

    def calculate_errors(self, data: np.ndarray, labels: np.ndarray) -> int:
        """
        Returns the divergence between two datasets using first nearest neighbor

        Parameters
        ----------
        data : np.ndarray
            Array containing images from two datasets
        labels : np.ndarray
            Array showing which dataset each image belonged to

        Returns
        -------
        int
            Number of edges connecting the two datasets
        """
        nn_indices = self._compute_neighbors(data, data)
        errors = np.sum(np.abs(labels[nn_indices] - labels))
        return errors
